package if688.plplus.analysis;

import if688.plplus.node.*;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import com.sun.xml.internal.bind.v2.runtime.reflect.ListIterator;

public class SemanticAnalyzer extends DepthFirstAdapter {

	Vector<Hashtable<String, Attribute>> symbolTable = new Vector<Hashtable<String, Attribute>>();
	
	public Hashtable<String, Attribute> escopoAtual(){
		return this.symbolTable.elementAt(symbolTable.size() - 1); 
	}
	
	public boolean buscaNosEscopos(String key){
		boolean retorno = false;
		
		int escopoDaVez = symbolTable.size() - 1;
		
		while(escopoDaVez >= 0){
			Hashtable<String, Attribute> escopoAtual = this.symbolTable.elementAt(escopoDaVez);
			if(escopoAtual.containsKey(key)){
				retorno = true;
				break;
			}
			escopoDaVez--;
		}
		
		return retorno;
	}
	
	public Attribute buscarAtributosDoIdentificador(String key){
		Attribute retorno = null;
		
		int escopoDaVez = symbolTable.size() - 1;
		
		while(escopoDaVez >= 0){
			Hashtable<String, Attribute> escopoAtual = this.symbolTable.elementAt(escopoDaVez);
			if(escopoAtual.containsKey(key)){
				retorno = escopoAtual.get(key);
				break;
			}
			escopoDaVez--;
		}
		
		return retorno;
	}
	
	public void abrirEscopo(){
		symbolTable.add(new Hashtable<String, Attribute>());
	}
	
	public void fecharEscopo(){
    	int tamanhoSymbolTable = symbolTable.size();
    	symbolTable.removeElementAt(tamanhoSymbolTable - 1);
	}
	
	public void inAModule(AModule node){
		this.abrirEscopo();		
    }

    public void outAModule(AModule node){
    	this.fecharEscopo();
    }
	
//    public void inAFunctiondecl(AFunctiondecl node){
//    	this.abrirEscopo();
//    }
    
    public void outAFunctiondecl(AFunctiondecl node){
    	this.fecharEscopo();
    }
    
//    public void inAProceduredecl(AProceduredecl node){
//    	this.abrirEscopo();
//    }
    
    public void outAProceduredecl(AProceduredecl node){
    	this.fecharEscopo();
    }
    
    public void inAProcedureheading(AProcedureheading node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador do procedimento " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		this.escopoAtual().put(key, new Attribute(Attribute.PROC));
    	}
    	this.abrirEscopo();
    }
    
    public void inAFunctionheading(AFunctionheading node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador da funo " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		Attribute atributos = new Attribute(node.getType().toString(), Attribute.FUNC);
    		this.escopoAtual().put(key, atributos);
    	}
    	this.abrirEscopo();
    	//criar variavel com o mesmo nome da funao
    	this.escopoAtual().put(key, new Attribute(node.getType().toString(), Attribute.VAR));
    }
	
    public void inAIdentlist(AIdentlist node) {
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		this.escopoAtual().put(key, new Attribute());
    	}
    }
    
    public void outAVardecl(AVardecl node){
    	AIdentlist identlist = (AIdentlist)node.getIdentlist();
    	String key = identlist.getIdent().getText().toUpperCase();
    	Attribute atributoAtual = this.escopoAtual().get(key);
    	atributoAtual.setType(node.getType().toString());
    	atributoAtual.identType = Attribute.VAR;
    	LinkedList<PSeqIdentlist> s = identlist.getSeqIdentlist();
    	Iterator<PSeqIdentlist> iterator = s.iterator();
    	PSeqIdentlist atual;
    	while(iterator.hasNext()){
    		 atual = iterator.next();
    		 ASeqIdentlist atualMesmo = (ASeqIdentlist)atual;
    		 key = atualMesmo.getIdent().getText().toUpperCase();
    		 atributoAtual = this.escopoAtual().get(key);
    		 atributoAtual.setType(node.getType().toString());
    		 atributoAtual.identType = Attribute.VAR;
    	}
    }
    
    public void inASeqIdentlist(ASeqIdentlist node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		this.escopoAtual().put(key, new Attribute());
    	}
    }
    
    public void inAFormalparamlist(AFormalparamlist node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		this.escopoAtual().put(key, new Attribute(node.getType().toString(), Attribute.VAR));
    	}
    }
    
    public void inASeqFormalparamlist(ASeqFormalparamlist node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(this.escopoAtual().containsKey(key)){
    		System.out.println("Identificador " + ident + " j existe no escopo");
    		System.exit(0);
    	}else{
    		this.escopoAtual().put(key, new Attribute(node.getType().toString(), Attribute.VAR));
    	}
    }
    
    public void inAAtribuicaoStatement(AAtribuicaoStatement node) {
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(!this.buscaNosEscopos(key)){
    		System.out.println("Identificador " + ident + " nao foi declarado!");
    		System.exit(0);
    	}
    }
    
    public void inAChamadaFuncaoStatement(AChamadaFuncaoStatement node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(!this.buscaNosEscopos(key)){
    		System.out.println("Identificador " + ident + " nao foi declarado!");
    		System.exit(0);
    	}
    }
    
    public void inAFatorFactor(AFatorFactor node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(!this.buscaNosEscopos(key)){
    		System.out.println("Identificador " + ident + " nao foi declarado!");
    		System.exit(0);
    	}
    }
    
    public void outAFatorFactor(AFatorFactor node){
    	Attribute att = this.buscarAtributosDoIdentificador(node.getIdent().toString());
    	if(att.identType == Attribute.FUNC){
        	node.setTipo(att.getTipo());
    	}else{
    		System.out.println(node.getIdent().toString() + " no  uma funo");
    	}
    }
    
    public void inAIdent9Factor(AIdent9Factor node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(!this.buscaNosEscopos(key)){
    		System.out.println("Identificador " + ident + " nao foi declarado!");
    		System.exit(0);
    	}
    }
    
    public void outAIdent9Factor(AIdent9Factor node){
        node.setTipo(this.buscarAtributosDoIdentificador(node.getIdent().toString()).getTipo());
    }
    
    public void inALacoForStatement(ALacoForStatement node){
    	TIdent ident = node.getIdent();
    	String key = ident.getText().toUpperCase();
    	
    	if(!this.buscaNosEscopos(key)){
    		System.out.println("Identificador " + ident + " nao foi declarado!");
    		System.exit(0);
    	}
    }
    
    public void outAIntLiteral9Factor(AIntLiteral9Factor node){
        node.setTipo(Attribute.INTEGER);
    }
    
    public void outACharLiteral0Factor(ACharLiteral0Factor node){
    	node.setTipo(Attribute.CHAR);
    }
    
    public void outABooleanLiteral9Factor(ABooleanLiteral9Factor node){
    	node.setTipo(Attribute.BOOLEAN);
    }
    
    public void outAFatorUnicoFactor(AFatorUnicoFactor node){
    	node.setTipo(((AExpression)node.getExpression()).getTipo());
    }
    
}
